###########################################################################
# cedl_scr.r                                                              #
#                                                                         #
# This script file contains the functions to implement Goodness of Fit    #
#  tests for the G and K statistics against a given H_0 based on the      #
#  Complete Empirical Distance List (CEDL).                               #
# The tests use MC methods to simulate patterns from the hypothesized     #
#  distribution, and then implement a Cramer von-Mises test statistic to  #
#  develop a null distribution which describes in general how simulated   #
#  patterns differ from the mean response.  Results for an observed       #
#  pattern are similarly compared to the mean response and the rank of    #
#  the observed result within the simlated results determines the p-value.#
#                                                                         #
# The general algorithm is as follows:                                    #
#  1.) Create s patterns of the specified null distribution.  These are   #
#   stored in the external my.pp.list variable.                           #
#  2.) Find the complete empirical distance lists for the obsrved pattern #
#   and each of the simulated patterns, and use these to process each     #
#   pattern individually with the statistic of choice.  Store the results #
#   in the hold variable, a list where each element has the results for a #
#   single pattern.  Each element of the hold list also contains the      #
#   pattern number, necessary for tracking which u_i value to assign its  #
#   results to.                                                           #
#  3.) Flatten the hold list by combining the elements into a single      #
#   matrix, and sort it by distance.                                      #
#  4.) Calculate the ui results for each pattern. This is done by walking #
#   through the hold list and at distance updating each u_i value. First, #
#   at each distance, the current K_i value is updated, as are the mean K #
#   values.  Then, the incremental summary statistic for each pattern is  #
#   calculated and added to the running total.  This is performed up to   #
#   the specified t_max value (the upper limit of integration).  The      #
#   vector of u_i, the complete empiricaldistance list, and the mean      #
#   result across all patterns, are returned.                             #
#  5.) Calculate the p-value of the observed pattern based on the rank of #
#   u_1 (by convention) against the remaining u_i.                        #
#                                                                         #
# Created 10/21/04 out of the complete.distlist_scr.r script file, but    #
#  now incorporates feedback from my visit to Diggle, et al.              #
# Updated 11/09/04 to add comments and documentation, and to make the G   #
#  and K processes more similar.  As of now, both create an external pp   #
#  list and then use this to run from.  This is to allow users to supply  #
#  their own pp list if they want to generate a corresponding simulation  #
#  envelope or create non-CSR patterns.                                   #
#                                                                         #
###########################################################################

# this script relies on functions from the spatstat library
require(spatstat)

###########################################################################
# General helper functions for both G and K                               #
###########################################################################

###########################################################################
# gen.ext.pp.list() generates an external CSR point pattern list.  Note,  #
#  this function overwrites any externally named variables `my.pp.list'   #
#  and `pp.indx' if they already exist.                                   #
#                                                                         #
# input variables:                                                        #
#  - nsim.pp is the number of simulated point patterns in this list       #
#  - npts is the number of points in each pattern                         #
#  - type is the pattern type to be generated                             #
#                                                                         #
# output variables:                                                       #
#  - ec is intended to check if the pattern list already exists (not yet  #
#  implemented)                                                           #
#                                                                         #
# created: 06/08/04                                                       #
# updated: 11/2/04 to allow a different null distribution to be used, but #
#  this still needs to update this to allow the text of the pattern type  #
#  to be used.                                                            #
###########################################################################
gen.ext.pp.list <- function(nsim.pp, npts, type="CSR")
{
	# create a blank list at the top level environment
	my.pp.list <<- c() 	# should be equivalent to <<- NULL
	# now popoulate it with the appropriate patterns.
	for (i in 1:nsim.pp)
		if (type=="CSR")
			my.pp.list[[i]] <<- runifpoint(npts)
		else    # for now, just have an inhibition option
			my.pp.list[[i]] <<- rSSI(r=0.0045, n=npts,
				giveup=10000)

	# lastly, create the top level index to the pattern list
	pp.indx <<- 1
	# nothing to return because the list is created externally.
	return(0)
}

###########################################################################
# get.next.pp() fetches the next point pattern (as indexed by pp.indx) in #
#  the external point pattern list.  It makes external reference to both. #
#                                                                         #
# output variables:                                                       #
#  - this.pp is the next pattern in the pattern list                      #
#                                                                         #
# created: 06/08/04                                                       #
###########################################################################
get.next.pp <- function()
{
	# locally define this pattern
	this.pp <- my.pp.list[[pp.indx]]
	# increment the index
	pp.indx <<- pp.indx + 1
	# and then return the pattern
	return(this.pp)
}

###########################################################################
# set.pp.prec() sets the precision of the x,y locations of the given      #
#  pattern to the appropriate level.                                      #
#                                                                         #
# input variables:                                                        #
#  - my.pp is the patten to be converted.                                 #
#  - prec is the desired precision of the ultimate result                 #
#                                                                         #
# output variables:                                                       #
#  - my.pp is an object of *.pp with x and y locations updated.           #
###########################################################################
set.pp.prec <- function(my.pp, prec=0.00001)
{
	# convert the precision into a number of digits value
	my.dig <- -log10(prec)
	# round the x and y location values
	my.pp$x <- round(my.pp$x, digits=my.dig)
	my.pp$y <- round(my.pp$y, digits=my.dig)
	# return the modified point pattern
	return(my.pp)
}

###########################################################################
# calc.dib() calculates the minimum distance from each point to the       #
#  boundary assuming a unit square.                                       #
#                                                                         #
# input variables:                                                        #
#  - my.pp is the patten from which to extract the d_ib values.           #
#                                                                         #
# output variables:                                                       #
#  - d.ib is the list of d_ib values for the points in the pattern        #
###########################################################################
calc.dib <- function(my.pp)
{
	# extract the x and y locations for each point
	x <- my.pp$x
	y <- my.pp$y
	d.ib <- pmin(x,y,1-x,1-y)
	# return the results
	return(d.ib=d.ib)
}

###########################################################################
# flatten.hold.list() called by cedl.G() and create.K.null.from.hold() to #
#  both flatten and sort the hold list.  Also, it only returns those      #
#  values less than tmax                                                  #
#                                                                         #
# input variables:                                                        #
#  - hold is a list with the results for each pattern                     #
#  - tmax is the largest distance to include within the resulting matrix  #
#                                                                         #
# output variables:                                                       #
#  - hold is the flattened (single array) and sorted (by distance) matrix #
#  containing the results for all of the patterns.                        #
#                                                                         #
# created 06/08/04                                                        #
###########################################################################
flatten.hold.list <- function(hold, tmax)
{
	# convert the hold list into a vector
	hold <- unlist(hold)
	# figure out the total number of distances
	ndist <- length(hold)/3
	# and then reconfigure it as a matrix which is 3 x ndist
	dim(hold) <- c(3, ndist)

	# now sort the data by distance, by keep in matrix form.
	hold <- hold[,order(hold[2,])]

#	# and re-round the distance list here, just to be sure...
#	hold[2,] <- round(hold[2,], digits=5)

	# now, include only those results that are <= to tmax
	hold <- hold[,hold[2,]<=tmax]

	return(hold)
}

###########################################################################
# calc.pval() takes in a distribution of ui and spits back a p-value for  #
#  the observed pattern (u1).  The p-value is the rank of the observed ui #
#  within the simulated ui.                                               #
#                                                                         #
# input variables:                                                        #
#  - ui is the list of calculated ui from either a G or K null distrib    #
#                                                                         #
# output variables:                                                       #
#  - my.pval is the p-value of the observed pattern (it must be in u_1!)  #
#                                                                         #
# created 11/09/04                                                        #
###########################################################################
calc.pval <- function(ui)
{
	# fetch the ui for the observed pattern
	my.ui <- ui[1]
	# calculate the number of values greater than the observed
	my.pval <- (length(ui[ui>=my.ui]))/(length(ui))

	return(my.pval)
}

###########################################################################
# G specific helper functions                                             #
###########################################################################

###########################################################################
# nndist.prec() determines the nearest neighbors distances for a given    #
#  pattern (actually using nndist), but limits those results by the       #
#  specified precision.  The rounding always occurs upwards, in order     #
#  that the subsequently calculated G result 'picks up' the intended      #
#  change at that distance.                                               #
#                                                                         #
# input variables:                                                        #
#  - my.pp is a point pattern which has already had x,y rounded           #
#  - prec is the desired precision of the ultimate result                 #
#                                                                         #
# output variables:                                                       #
#  - mysig.nn is a sorted unique distance list.                           #
###########################################################################
nndist.prec<- function(my.pp, prec=0.00001)
{
	# convert the precision into a number of digits value
	my.dig <- -log10(prec)

	# fetch the distance list (sorted just to make it easy to see).
	my.nn <- sort(unique(nndist(my.pp)))

	# round the distances to the same acceptable level of precision
	mysig.nn <- round(my.nn, digits=my.dig)

	# now look at those that were rounded down and add the precision
	# estimate.  Because of how the G statistic works, select the next
	# highest rounded distance.  I want to make sure the $\hat G$
	# value increases at the distance.
	mysig.nn[mysig.nn < my.nn] <- mysig.nn[mysig.nn < my.nn] + prec

	# return the nearest neighbor list
	return(mysig.nn)
}

###########################################################################
# calc.Gest.rslt() calculates the Gest result for a random pattern and    #
#  stores it in the rslt[] matrix.  'i' is the pattern number.            #
#                                                                         #
# input variables:                                                        #
#  - this.pp is the pattern to calculate the G function for               #
#  - i is the pattern number                                              #
#  - prec is the desired precision to use for point locations             #
#  - ec.meth is the edge correction method (not currently implemented)    #
#                                                                         #
# output variables:                                                       #
#  - results is a matrix with dimensions (3, number of distances) which   #
#  contains [1,] the G results for the pattern, [2,] the actual distances #
#  at which the G function is computed, and [3,] the pattern number.      #
#                                                                         #
# created 04/28/04 from the guts of comp.distlist.G.null()                #
###########################################################################
calc.Gest.rslt <- function(this.pp, i, prec, ec.meth="rs")
{
	# determine its actual nearest neighbor dists, again w/ precision
	this.nn <- c(0, nndist.prec(this.pp, prec), sqrt(2))
	
	# but if using reduced sample, add in the d.ib values
	if (ec.meth == "rs") {
		this.dib <- calc.dib(this.pp)
		this.nn <- sort(c(this.nn, this.dib))
	}

	# setup to store the results
	results <- matrix(0, 3, length(this.nn))
	# and fetch the results from the Gest based on the ec.method
	if (ec.meth == "rs") {
		results[1,] <- Gest(this.pp, r=this.nn)$rs
		# and then `fix' those locations where NaN
		results[1,][is.nan(results[1,])] <- 1
	} else
		# assume 'raw' results
		results[1,] <- Gest(this.pp, r=this.nn)$raw

	# store the distances and the pattern id
	results[2,] <- this.nn
	results[3,] <- i

	return(results)
}

###########################################################################
# calc.G.ui.step() is a helper function called by calc.G.ui to perform    #
#  the summary statistic calculation at a specific delta.t.  This is      #
#  where the actual Cramer von-Mises equation is implemented.             #
#                                                                         #
# input variables:                                                        #
#  - Gt is the vector of G results, one for each pattern, at this dist    #
#  - delta.t is the distance increment for this calculation               #
#  - num.pattern is the number of patterns within the null, including the #
#  obsvd.  Note this could probably be extracted from length(Gt).         #
#                                                                         #
# output variables:                                                       #
#  - ui.step is the vector of incremental ui values, one for each pattern #
#  - Gi.bar is the mean result across all patterns at this distance       #
#                                                                         #
# created 11/10/04 out of the calc.G.ui() innards to make the code a      #
#  little more readable.                                                  #
###########################################################################
calc.G.ui.step <- function(Gt, delta.t, num.pattern)
{
	# calculate the Gi.bar results for all patterns.  hence Gi.bar
	#  is a vector, but local to this if statement.
	Gi.bar <- (sum(Gt) - Gt)/(num.pattern - 1)

	# compute the value of ui for the latest step across all patterns
	#  note I'm not using ^2 here because it seems to takes longer than
	#  just multiplying.
	ui.step <- (Gt-Gi.bar)*(Gt-Gi.bar) * delta.t

	return(list(ui.step=ui.step, Gi.bar=mean(Gt)))
}

###########################################################################
# calc.G.ui() is a helper function used by cedl.G to calculate all of the #
#  ui values based on the flattened and sorted hold list.                 #
#                                                                         #
# input variables:                                                        #
#  - hold is the flattened (into a single array) and sorted (by distance) #
#  matrix (3 x ndist) containing the G results for all of the patterns.   #
#  - num.patterns is the total number of patterns (including both the     #
#  observed and simulated) within the hold array                          #
#                                                                         #
# output variables:                                                       #
#  - ui is the unsorted list of u_i values                                #
#  - Go is the mean Go value calculated across all of the patterns        #
#  - my.dist is the list of distances at which Go was calculated.         #
#                                                                         #
# updated 4/15/04 to deal with the precision issue. given limited         #
#  precision its possible that two patterns will have results at the same #
#  `t'.  This function then needs to look out for and account for this    #
#  situation.                                                             #
# updated 10/19/04 since if the observed pattern is included, we need to  #
#  know what the first value is (so don't sort it).                       #
# updated 11/10/04 as part of the commenting process to clean up the main #
#  loop.                                                                  #
###########################################################################
calc.G.ui <- function(hold, num.pattern)
{
	# create storage for the G(t) per pattern at the given t value. 
	#  This is updated only for the current pattern.
	Gt <- vector("numeric", num.pattern)
	
	# create storage for the ui value per pattern.  This is updated
	#  at each unique distance increment, because something must have
	#  changed.
	ui <- vector("numeric", num.pattern)
	
	# create storage for the mean (across all patterns) G response
	# I don't know its length because of duplicates, so instead use
	#  unique(), to find the number of unique distances 
	nuniq.dist <- length(unique(hold[2,]))
	Go <- vector("numeric", nuniq.dist)
	# and then use internal counter (g.indx) to index it
	g.indx <- 1

	# create storage for the actual distances at which Go is calc'd
	my.dist <- vector("numeric", nuniq.dist)

	# main for loop to run through each distance and calculate
	#  the incremental change to ui.  This length is the max
	#  I might loop across, but I'm going to break out of it once
	#  the calculated mean value Go=1, because then I'm done.
	for (t in 1:(length(hold[2,])-1)) {
		# extract the current pattern id
		this.pat <- hold[3,t]
		# update the current Gt vector with this distance's result
		Gt[this.pat] <- hold[1,t]
		# determine the distance increment
		delta.t <- (hold[2, t+1]-hold[2, t])

	# Now check if this distance is the same as the next one, and if so
	# skip this next part because we need to further updated Gt[]
		if (delta.t != 0) {
		# calculate the ui.step for this delta.t
			t.rslt <- calc.G.ui.step(Gt, delta.t, num.pattern)

		# update the running ui value for each pattern
			ui <- ui + t.rslt$ui.step

		# store the overall mean (across all patterns) for this t
			Go[g.indx] <- t.rslt$Gi.bar
	
		# and store the distance associated with this Go value
			my.dist[g.indx] <- hold[2, t]

	# now check if we're done, this is just a short cut to make things
	#  run a little faster, since if the mean equals 1, all subsequent
	#  values will also be 1.
			if (Go[g.indx]==1) {
				break
			}
	# update the Go index variable so we can address the next element 
			g.indx <- g.indx+1
		}
	}

	# now, get rid of those points where the distance is zero
	Go <- Go[my.dist !=0]
	my.dist <- my.dist[my.dist !=0]

	# and then make sure the first and last values are set right
	Go <- c(0, Go, 1)
	# note: this isn't quite right, and should instead use t_max
	#  and/or the prec variables.
	my.dist <- c(0, my.dist, round(sqrt(2), 5))
	
	return(list(ui=ui, Go=Go, my.dist=my.dist))
}

###########################################################################
# gen.G.hold.list() generates a hold list containing the G statistic      #
#  results for a given number of patterns.  it assumes that the pattern   #
#  list has already been created and exists in an external my.pp.list     #
#  variable.                                                              #
#                                                                         #
# input variables:                                                        #
#  - obs.pp is the observed pattern                                       #
#  - prec is the desired precision to set both the observed and simualted #
#  patterns to.                                                           #
#  - verbose is a boolean used to allow progress indication               #
#  - ec.meth is the edge correction method to use for the G statistic     #
#                                                                         #
# output variables:                                                       #
#  - hold is the list where each element contains the results for one of  #
#  the patterns.  the first element is the observed pattern and the rest  #
#  are the simulated ones.                                                #
#                                                                         #
# created: 11/09/04                                                       #
###########################################################################
gen.G.hold.list <- function(obs.pp, prec, ec.meth, verbose)
{
	hold <- NULL
	# extract the number of patterns
	nsim.pp <- length(my.pp.list)

	# process the observed pattern
	obs.pp <- set.pp.prec(obs.pp, prec)
	results <- calc.Gest.rslt(obs.pp, i=1, prec, ec.meth)

	# and start a counter to track the cumulative number of distances
	ndist <- length(results[1,])
	hold[[1]] <- results

	# now create all of the random patterns in the null distribution
	for (i in 2:(nsim.pp+1)) {
		if (verbose) cat(".")
	# create this a random pattern with the given intensity, and
	# update this pattern with the correct precision
		this.pp <- set.pp.prec(get.next.pp(), prec)

	# calculate the Gest results for the given pattern
		results <- calc.Gest.rslt(this.pp, i, prec, ec.meth)

	# store the number of distances over which Gest is calculated
		ndist <- ndist + length(results[1,])

	# and add this to the existing results
		hold[[i]] <- results
	}
	if (verbose) cat("\n")
	return(hold)
}

###########################################################################
# cedl.G()                                                                #
# the point of this function is to build a Null distribution based on the #
#  `complete' empirical distance list.  This consist of                   #
#  i) the mean G(t) result across the observed pattern and all sims,      #
#  ii) the list of distances where G(t) is calculated,                    #
#  iii) the vector of the calculated summary statistics, u_i              #
#                                                                         #
# the goal of this algorithm was to be more efficient with memory here,   #
#  at the expense of speed.  So, instead of storing a large (s by t)      #
#  matrix, only store a (3 by t) matrix, where t is the complete distance #
#  list.  this vector has three rows to include (G(t), t, pattern.id)     #
# then, when processing, go through each distance in t, and calcuate the  #
#  running G(t) values for each of the patterns.  So, basically, I'm only #
#  creating an (s by 1) vector, but requiring to loop through each        #
#  distance.                                                              #
#                                                                         #
# input variables:                                                        #
#  - obs.pp is the observed pattern to be evaluated                       #
#  - s is the number of patterns used to create the null distrib          #
#  - lambda is the number of point in each pattern                        #
#  - tmax is the maximum distance up to which the integration should occur#
#  - prec is the desired precision of the locations of simulated patterns #
#  - use.exist is a boolean to allow an existing external pattern list to #
#  be used, instead of generating a new one.                              #
#  - verbose is a boolean to allow progress messages to be reported       #
#                                                                         #
# output variables:                                                       #
#  - p.val is the simulated p-value of the observed pattern               #
#  - ui is the unsorted list of u_i values                                #
#  - Go is the mean Go value calculated across all of the patterns        #
#  - my.dist is the list of distances at which Go was calculated.         #
#                                                                         #
# Updated 4/15/04 to include precision in the locations and distances     #
# Updated 4/26/04 to rename to be G specific                              #
# Updated 10/19/04 to rename to cedl.G.null and to have the observed      #
#  pattern be part of the null distribution.   Also, renamed to `s' and   #
#  `lambda'                                                               #
# Updated 11/09/04 to generate and then use an external pattern list.     #
#  Also changed to use flatten.hold.list() similar to the cedl.K function.#
###########################################################################
cedl.G <- function(obs.pp, s, lambda=obs.pp$n, tmax=1, prec=0.00001,
	use.exist=F, verbose=F)
{
	# only runs using the $rs method
	ec.meth="rs"
	###################################################################

	if (!use.exist) {
		# create the pattern list used to create the null
		if (verbose) cat("generating patterns\n")
		# in this case, use CSR patterns
		ec <- gen.ext.pp.list(s-1, lambda, type="CSR")
		if (ec) return()
	}
	else {
		if (verbose) cat("using existing pattern list\n")
		# reset the top level index into the pattern list
		pp.indx <<- 1
	}

	if (verbose) cat("creating hold list\n")
	# calculate the hold list using the results from the observed and
	#  simulated patterns
	hold <- gen.G.hold.list(obs.pp, prec, ec.meth, verbose)

	###################################################################
	if (verbose) cat("creating null distrib\n")

	# changed 11/09/04 to use the flatten.hold.list() function
	hold <- flatten.hold.list(hold, tmax)

	# call calc.ui to calculate the ui result for each pattern and the
	#  mean Go result for each t
	ui.rslt <- calc.G.ui(hold, s)

	# remove the redundant distances where distance ==0 or == sqrt(2)
	# my.dist <- hold[2,][(hold[2,] >0 & hold[2,]<sqrt(2))]
	if (verbose) cat("\n")
	p.val <- calc.pval(ui.rslt$ui)
	if (verbose) cat("P-value:", p.val, "\n")
	# return the null distribution results
	return(list(p.val=p.val, ui=ui.rslt$ui, Go=ui.rslt$Go,
		 my.dist=ui.rslt$my.dist))
}

###########################################################################
# K specific helper functions                                             #
###########################################################################

###########################################################################
# andist.prec() computes all neighbor distances for a given pattern,      #
#  based on the precision value.                                          #
#                                                                         #
# input variables:                                                        #
#  - my.pp is a point pattern which has already had x,y rounded           #
#  - prec is the desired precision of the ultimate result                 #
#                                                                         #
# output variables:                                                       #
#  - mysig.nn is a sorted unique distance list.                           #
###########################################################################
andist.prec <- function(my.pp, prec=0.00001)
{
	# convert the precision into a number of digits value
	my.dig <- -log10(prec)

	# fetch the complete (all) neighbor distance list.
	my.an <- pairdist(my.pp)

	# round the distances to the same acceptable level of precision
	mysig.an <- round(my.an, digits=my.dig)

	# now look at those that were rounded down and add the precision
	# estimate.  Because of how the K statistic works, select the next
	# highest rounded distance.  I want to make sure the $\hat K$ value
	# increases at the distance.
	mysig.an[mysig.an < my.an] <- mysig.an[mysig.an < my.an] + prec 

	# return the all neighbor list
	return(mysig.an)
}

###########################################################################
# calc.Kest.rslt() calculates the Kest result for a random pattern        #
# `i' is the pattern number which is stored in the resultant matrix.      #
#  This function accomodates the precision issue.                         #
#                                                                         #
# input variables:                                                        #
#  - this.pp is the pattern to be analyzed                                #
#  - i is the pattern number                                              #
#  - prec is the precision to use for this pattern                        #
#  - use.L is a boolean as to whether the L statistic should be used in   #
#  place of K                                                             #
#                                                                         #
# output variables:                                                       #
#  - results is a matrix with dimensions (3, number of distances) which   #
#  contains [1,] the G results for the pattern, [2,] the actual distances #
#  at which the G function is computed, and [3,] the pattern number.      #
#                                                                         #
# updated 04/24/04 to add NA as the one after the last location           #
# updated 04/28/04 to take this.pp as a paramater instead of npts.  This  #
#  allows it to be called in test.K.pp().  It does not assume this.pp has #
#  already set the appropriate precision.                                 #
# updated 05/24/04 to allow the use of L instead of K.                    #
###########################################################################
calc.Kest.rslt <- function(this.pp, i, prec, use.L=F)
{
#	print("generating distance list")

	# extract the number of points in the pattern
	npts <- this.pp$n
	# determine the distances from each point to ALL of its neighbors,
	full.an <- andist.prec(this.pp, prec)
	# determine the distance from each point to the edge of the box
	this.dib <- calc.dib(this.pp)
	# then, for each point, include only those distances which are less
	#  than the dib value.  Note: this is ec="border" specific.
	# Also: this might be faster using tapply with a custom function
	limit.an <- full.an
	for (j in 1:npts) {
		limit.an[j,][limit.an[j,]>this.dib[j]] <- 0
	}

	# now convert to a vector and add in the dib values
	this.an <- c(as.vector(limit.an), this.dib)
	# then fetch only the unique values and sort
	this.an <- sort(unique(this.an))

#	print("calculating Kest")
	# setup to store the results from this pattern
	results <- matrix(0, 3, length(this.an))
	# and fetch the results from the Gest based on the ec.method
	results[1,] <- Kest(this.pp, r=this.an, correction="border")$border
	# store the distances and the pattern id
	results[2,] <- this.an
	results[3,] <- i

	# check if using the L transformation and if so, store this
	if (use.L) {
		results[1,] <- sqrt(results[1,]/pi)-results[2,]
	}

	# and now set the last value to `NA' as a place marker for the end
	#  of the results.  Note: its ok to use the last location since we
	#  wont ever calculate anything based on it.  Rem that K(t) is a
	#  step function
	results[1,length(this.an)] <- NA

	return(results)
}

###########################################################################
# gen.K.hold.list() generates a hold list containing the K statistic      #
#  results for a given number of patterns.  it assumes that the pattern   #
#  list has already been created and exists in an external my.pp.list     #
#  variable.                                                              #
#                                                                         #
# input variables:                                                        #
#  - obs.pp is the observed pattern                                       #
#  - prec is the desired precision to set both the observed and simualted #
#  patterns to.                                                           #
#  - verbose is a boolean used to allow progress indication               #
#                                                                         #
# output variables:                                                       #
#  - hold is the list where each element contains the results for one of  #
#  the patterns.  the first element is the observed pattern and the rest  #
#  are the simulated ones.                                                #
#                                                                         #
# created: 06/08/04                                                       #
###########################################################################
gen.K.hold.list <-function(obs.pp, prec, verbose)
{
	# determine the number of patterns in the external pattern list
	nsim.pp <- length(my.pp.list)

	hold <- c()

	# set the precision of the observed pattern
	obs.pp <- set.pp.prec(obs.pp, prec)
	# store the Kest results and store
	hold[[1]] <- calc.Kest.rslt(obs.pp, i=1, prec)

	for (i in 2:(nsim.pp+1))  {
		if (verbose) cat(".")
		# fetch the next pattern and set its precision
		this.pp <- set.pp.prec(get.next.pp(), prec)
		# calculate the Kest matrix for this pattern
		hold[[i]] <- calc.Kest.rslt(this.pp, i, prec)
	}
	if (verbose) cat("\n")
	return(hold)
}


###########################################################################
# calc.K.ui.step() calculates the ui.step value for the current Kit       #
#  vector. It is called within the for loop of create.K.null.from.hold(). # 
#  It returns the vector of ui.step. the mean value Kt.bar for this       #
#  distance, and the numbr of patterns comprising the mean at this dist.  #
# this function does NOT check whether t is less than tmax, that should   #
#  be done outside of this function.  in fact, it doesn't know what the   #
#  value of t is, only delta.t                                            #
#                                                                         #
# input variables:                                                        #
#  - Kit is the list of current Kest values for each pattern (i) at the   #
#  given distance (t)                                                     #
#  - delta.t is the distance bucket width, needed to calculate ui.step    #
#  - npp is the number of patterns                                        #
#                                                                         #
# output variables:                                                       #
#  - ui.step is the list of u_i values for the specified distance and     #
#  delta.t                                                                #
#  - Kit.bar is the mean value of K calculated across the valid patterns  #
#  - nvalid is the number of valid observations at this distances         #
#                                                                         #
# created 06/08/04, pulled from the internals of calc.K.ui2()             #
###########################################################################
calc.K.ui.step <- function(Kit, delta.t, npp)
{
	# make the c parameter easily changeable
	c <- 0.25
	
	# extract only those patterns that are valid.  
	valid <- (1:npp)[!is.na(Kit)]
	# and store that total number 
	nvalid <- length(valid)

	# extract a matrix of the values of those that are valid
	Kt.val <- Kit[valid]
	
	# calculate the Ki.bar results (mean excluding pattern i) across
	#  all valid patterns, hence Ki.bar has length of nvalid
	if (nvalid > 1)
		Ki.bar <- (sum(Kt.val) - Kt.val)/(nvalid-1)
	# now if there is only one valid pattern, its mean (excluding it)
	#  is obviously zero!  
	else
		Ki.bar <- 0

	# compute the value of ui for the latest step across all patterns.  
	#  This equation is based on Diggle (2003) equation 6.5
	# Not using ^2 here because the calculation takes much longer.
	this.ui <- (Kt.val^c-Ki.bar^c) *
		(Kt.val^c-Ki.bar^c) * delta.t

	# create the vector to store the results and initialize it to zero
	#  so that non valid results will remain zero.
	ui.step <- vector("numeric", npp)
	# add the results to those valid locations
	ui.step[valid] <- ui.step[valid] + this.ui
	
	# calculate the overall mean (across all valid patterns) 
	Kit.bar <- mean(Kt.val)

	return(list(ui.step=ui.step, Kit.bar=Kit.bar, nvalid=nvalid))
}

###########################################################################
# create.K.null.from.hold() creates a null distrib from the hold list.    #
#  This implements most of the functionality of calc.K.ui2() but using an #
#  alternative breakdown of the internal functions                        #
#                                                                         #
# input variables:                                                        #
#  - hold is the hold list, where each element contains the results for a #
#  single pattern.  This list is not yet flattened nor sorted.            #
#  - tmax is the maximum distance across which integration should occur   #
#                                                                         #
# output variables:                                                       #
#  - ui is the list of u_i for all patterns                               #
#  - Ko is a matrix (3x num unique distances) which contains [1,] the     #
#  mean Ko result across those patterns that are valid, [2,] the dists    #
#  at which the Ko value is computed, and [3,] the num of patterns that   #
#  are valid at the given distance and therefore used in computing Ko.    #
#                                                                         #
# created: 06/08/04                                                       #
###########################################################################
create.K.null.from.hold <- function(hold, tmax)
{
	# flatten, sort and shorten (up to tmax) the hold list
	hold <- flatten.hold.list(hold, tmax)
	# and extract the number of patterns
	npp <- length(unique(hold[3,]))
	
	###################################################################
	#### setup all of the internal variables                       ####
	
	# the Kest results at this distance for each pattern
	Kit <- vector("numeric", npp)
	# the running results of u_i for each pattern up to this distance
	ui <- vector("numeric", npp)

	# the numer of unique distances within the hold matrix
	nuniq.dist <- length(unique(hold[2,]))-1
	# storage for the mean Ko.bar result at each distance,
	Ko <- matrix(0, 3, nuniq.dist)
	
	# kindx is used to index the results at each unique distance
	kindx <- 0
	# ntot.dist is the number of total distances to loop across
	ntot.dist <- length(hold[2,])-1
	
	###################################################################
	#### main for loop to evaluate the hold vector                 ####
	for (t in 1:ntot.dist) {
#		cat(".")
		# update the results for the given pattern at this distance
		Kit[hold[3,t]] <- hold[1,t]
		# determine the distance increment
		delta.t <- (hold[2, t+1]-hold[2, t])

		# now, if delta.t is zero, then we need to update the Kit
		#  vector more before evaluating.
		# Otherwise, we have results to process, so go forward
		if (delta.t >0) {
			# increment the kindx variable
			kindx <- kindx +1
			# perform the calculation on this Kit list
			t.rslt <- calc.K.ui.step(Kit, delta.t, npp)
			# update the ui results,
			ui <- ui + t.rslt$ui.step
			# and the Ko results
			Ko[, kindx] <- c(t.rslt$Kit.bar, hold[2, t],
				t.rslt$nvalid)
		}
	}
	# and return the entire results.
#	cat("\n")
	return(list(ui=ui, Ko=Ko))
}

###########################################################################
# cedl.K() creates a null distribution for the K statistic and tests an   #
#  observed pattern against it.                                           #
# the point of this function is to build a Null distribution based on the #
#  `complete' empirical distance list.  This has three parts:             #
#  i) the mean K(t) result across all simulations,                        #
#  ii) the list of distances where K(t) is calculated,                    #
#  iii) the `sorted' vector of the calculated summary statistics, u_i     #
#                                                                         #
# input variables:                                                        #
#  - obs.pp is the observed pattern to be evaluated                       #
#  - s is the number of patterns used to create the null distrib          #
#  - lambda is the number of point in each pattern                        #
#  - tmax is the maximum distance up to which the summation should occur  #
#  - prec is the desired precision of the locations of simulated patterns #
#  - use.exist is a boolean to allow an existing external pattern list to #
#  be used, instead of generating a new one.                              #
#  - verbose is a boolean to allow progress messages to be reported       #
#                                                                         #
# output variables:                                                       #
#  - p.val is the simulated p-value of the observed pattern               #
#  - ui is the unsorted list of u_i values                                #
#  - Ko is the mean Ko value calculated across all of the patterns        #
#  - my.dist is the list of distances at which Go was calculated.         #
#                                                                         #
# created 10/20 from create.K.null() to include the observed pattern in   #
#  the null and use the new (consistent) function header.                 #
# updated 11/09 to change the function call, adding a default value for   #
#  lambda and allowing the existing pp list to be used.                   #
###########################################################################
cedl.K <- function(obs.pp, s, lambda=obs.pp$n, tmax=0.25, prec=0.00001,
	 use.exist=F, verbose=F)
{
	if (!use.exist) {
		# create the pattern list used to create the null
		if (verbose) cat("generating patterns\n")
		# in this case, use CSR patterns
		ec <- gen.ext.pp.list(s-1, lambda, type="CSR")
		if (ec) return()
	}
	else {
		if (verbose) cat("using existing pattern list\n")
		# reset the top level index into the pattern list
		pp.indx <<- 1
	}

	# generate the hold list containing the actual Kest results for
	#  each pattern
	if (verbose) cat("creating hold list\n")
	hold <- gen.K.hold.list(obs.pp, prec, verbose)
	# develop the null distribution, particularly the u_i values, from
	#  this hold list
	if (verbose) cat("creating null distrib\n")
	K.rslt <- create.K.null.from.hold(hold, tmax)
	if (verbose) cat("\n")

	# calculate the p-value of the observed pattern
	p.val <- calc.pval(K.rslt$ui)
	if (verbose) cat("P-value:", p.val, "\n")

	# return the important results.  Note:  I'm disecting the K.rslt
	#  matrix, but discarding the nvalid data
	return(list(p.val=p.val, ui=K.rslt$ui, Ko=K.rslt$Ko[1,],
		 my.dist=K.rslt$Ko[2,]))
}

